#include "SineSynth.h"

#include <iostream>
#include <random>
#include <cassert>

template <typename dataType>
Generator<dataType>::Generator(uint32_t _sampleRate, float _frequency, float _amplitude)
    : sampleRate(_sampleRate), frequency(_frequency), amplitude(_amplitude), enabled(false)
{
}

template <typename dataType>
Generator<dataType>::~Generator() {}

template <typename dataType>
void Generator<dataType>::getSamples(size_t nSamples, dataType *buf)
{
    for (; nSamples > 0; nSamples--)
    {
        *buf++ = getSample();
    }
}

template <typename dataType>
float Generator<dataType>::getFrequency() const { return this->frequency; }

template <typename dataType>
void Generator<dataType>::setFrequency(float frequency) { this->frequency = frequency; }

template <typename dataType>
float Generator<dataType>::getAmplitude() const { return this->amplitude; }

template <typename dataType>
void Generator<dataType>::setAmplitude(float amplitude) { this->amplitude = amplitude; }

template <typename dataType>
bool Generator<dataType>::isEnabled() const { return enabled; }

template <typename dataType>
void Generator<dataType>::enable() { enabled = true; }

template <typename dataType>
void Generator<dataType>::disable() { enabled = false; }

template <typename dataType>
SineGenerator<dataType>::SineGenerator(
    uint32_t _sampleRate, float _frequency, float _amplitude)
    : Generator<dataType>(_sampleRate, _frequency, _amplitude)
{
    mpfr_init_set_ui(instPhase, 0, MPFR_RNDN);
    mpfr_init(phaseIncr);
    mpfr_init(PI_2);
    mpfr_init(tmpA);

    mpfr_const_pi(PI_2, MPFR_RNDN);
    mpfr_mul_ui(PI_2, PI_2, 2, MPFR_RNDN);

    setFrequency(_frequency);
    this->enable();
    std::cout << "Constructed SineGenerator with freq " << this->frequency
              << " phaseIncr " << mpfr_get_d(phaseIncr, MPFR_RNDN) << " rads/cycle" << std::endl;
}

template <typename dataType>
SineGenerator<dataType>::~SineGenerator()
{
    mpfr_clear(instPhase);
    mpfr_clear(phaseIncr);
    mpfr_clear(PI_2);
    mpfr_clear(tmpA);
}

template <>
float SineGenerator<float>::getSample()
{
    mpfr_add(instPhase, instPhase, phaseIncr, MPFR_RNDN);
    if (mpfr_cmp(instPhase, PI_2) > 0)
    {
        mpfr_sub(instPhase, instPhase, PI_2, MPFR_RNDN);
    }
    mpfr_cos(tmpA, instPhase, MPFR_RNDN);
    mpfr_mul_d(tmpA, tmpA, this->amplitude, MPFR_RNDN);

    return mpfr_get_flt(tmpA, MPFR_RNDN);
}

template <>
uint32_t SineGenerator<uint32_t>::getSample()
{
    mpfr_add(instPhase, instPhase, phaseIncr, MPFR_RNDN);
    if (mpfr_cmp(instPhase, PI_2) > 0)
    {
        mpfr_sub(instPhase, instPhase, PI_2, MPFR_RNDN);
    }
    mpfr_cos(tmpA, instPhase, MPFR_RNDN);
    mpfr_mul_d(tmpA, tmpA, this->amplitude, MPFR_RNDN);
    mpfr_mul_ui(tmpA, tmpA, (uint32_t)INT32_MAX+1, MPFR_RNDN);

    return (uint32_t)(mpfr_get_si(tmpA, MPFR_RNDN));
}

//phaseIncr = 2 * M_PI * this->frequency / this->sampleRate;
template <typename dataType>
void SineGenerator<dataType>::setFrequency(mpfr_t _frequency)
{
    mpfr_mul(phaseIncr, PI_2, _frequency, MPFR_RNDN);
    mpfr_div_ui(phaseIncr, phaseIncr, this->sampleRate, MPFR_RNDN);
    this->frequency = mpfr_get_flt(_frequency, MPFR_RNDN);
}

template <typename dataType>
void SineGenerator<dataType>::setFrequency(float _frequency)
{
    mpfr_mul_d(phaseIncr, PI_2, _frequency, MPFR_RNDN);             // 2*pi*frequency
    mpfr_div_ui(phaseIncr, phaseIncr, this->sampleRate, MPFR_RNDN); // 2*pi*frequency / samplerate
    this->frequency = _frequency;
}

template <typename dataType>
SineSynth<dataType>::SineSynth(uint32_t _sampleRate)
    : sampleRate(_sampleRate)
{
}

template <typename dataType>
void SineSynth<dataType>::addSynth(shared_ptr<Generator<dataType>> synth)
{
    generators.push_back(synth);
}

template <typename dataType>
vector<shared_ptr<Generator<dataType>>> SineSynth<dataType>::getSynths() const
{
    return generators;
}

template <typename dataType>
void SineSynth<dataType>::clearSynths() { generators.resize(0); }

template <typename dataType>
void SineSynth<dataType>::getSamples(size_t nSamples, dataType *buf)
{
    for (; nSamples > 0; nSamples--)
    {
        *buf = 0;
        for_each(std::begin(generators), std::end(generators),
                 [buf](std::shared_ptr<Generator<dataType>> g) {
                     if (g->isEnabled())
                         *buf += g->getSample();
                 });
        buf++;
    }
}

template <typename dataType>
NoiseGenerator<dataType>::NoiseGenerator(
    uint32_t _sampleRate, NOISE_TYPE type, float _amplitude)
    : Generator<dataType>(_sampleRate, 0, _amplitude), rand_gen()
{
    this->enable();
}

template <typename dataType>
NoiseGenerator<dataType>::~NoiseGenerator() {}

template <typename dataType>
dataType NoiseGenerator<dataType>::getSample()
{
    return std::uniform_real_distribution<float>(-this->amplitude, std::nextafter(this->amplitude, std::numeric_limits<float>::max()))(rand_gen);
}

template <typename dataType>
JTestGenerator<dataType>::JTestGenerator(uint32_t sampleRate, uint8_t nbits) : Generator<dataType>(sampleRate, 0, 1)
{
    size_t loopLength;
    if (sampleRate % 250 == 0)
        loopLength = sampleRate / 250;
    else if (sampleRate % 245 == 0)
        loopLength = sampleRate / 245;
    else
    {
        std::cerr << "Unable to generate non-integral fundamental frequency for JTest";
        exit(-1);
    }

    this->loop.resize(loopLength);

    // Build signal
    const auto shift = nbits - 8;
    const uint8_t posbase = 0x40;
    const uint8_t negbase = 0xc0;

    const auto posval_first = posbase << shift;
    const auto negval_first = negbase << shift;
    const auto posval_last = posval_first - 1;
    const auto negval_last = negval_first - 1;

    for (auto i = 0; i < loopLength / 2; i += 4)
    {
        this->loop[i + 0] = negval_first;
        this->loop[i + 1] = negval_first;
        this->loop[i + 2] = posval_first;
        this->loop[i + 3] = posval_first;
    }
    for (auto i = loopLength / 2; i < loopLength; i += 4)
    {
        this->loop[i + 0] = negval_last;
        this->loop[i + 1] = negval_last;
        this->loop[i + 2] = posval_last;
        this->loop[i + 3] = posval_last;
    }

    this->pos = this->loop.begin();
}

template <typename dataType>
dataType JTestGenerator<dataType>::getSample()
{
    if (this->pos == this->loop.end())
        this->pos = this->loop.begin();
    return *(this->pos++);
}

template <typename dataType>
JTestGenerator<dataType>::~JTestGenerator() {}

template <typename dataType>
SweepGenerator<dataType>::SweepGenerator(uint32_t sampleRate, float startFreq, float endFreq, float length, float amplitude) : Generator<dataType>(sampleRate, startFreq, amplitude),
                                                                                                                               sg(sampleRate, startFreq, amplitude)
{
    this->enable();
    mpfr_init(freqStep);
    mpfr_init_set_d(this->startFreq, startFreq, MPFR_RNDN);
    mpfr_init_set(curFreq, this->startFreq, MPFR_RNDN);
    mpfr_init_set_d(this->endFreq, endFreq, MPFR_RNDN);

    // Linear frequency step = ((endFreq - startFreq) / length) / sampleRate
    mpfr_sub(freqStep, this->endFreq, this->startFreq, MPFR_RNDN);
    mpfr_div_d(freqStep, freqStep, length, MPFR_RNDN);
    mpfr_div_ui(freqStep, freqStep, sampleRate, MPFR_RNDN);
}

template <typename dataType>
SweepGenerator<dataType>::~SweepGenerator()
{
    mpfr_clear(startFreq);
    mpfr_clear(curFreq);
    mpfr_clear(endFreq);
    mpfr_clear(freqStep);
}

template <typename dataType>
dataType SweepGenerator<dataType>::getSample()
{
    dataType sample = sg.getSample();
    mpfr_add(curFreq, curFreq, freqStep, MPFR_RNDN);
    if (mpfr_cmp(curFreq, endFreq) > 0)
    {
        mpfr_set(curFreq, startFreq, MPFR_RNDN);
    }
    sg.setFrequency(curFreq);
    return sample;
}

template <typename dataType>
float SweepGenerator<dataType>::getFrequency() const
{
    return mpfr_get_flt(curFreq, MPFR_RNDN);
}

template <typename dataType>
void SweepGenerator<dataType>::setAmplitude(float amplitude) {
    sg.setAmplitude(amplitude);
    this->amplitude = amplitude;
}

template class NoiseGenerator<float>;
template class SineGenerator<float>;
template class SweepGenerator<float>;
template class SineSynth<float>;

template class SineSynth<uint32_t>;
template class SineGenerator<uint32_t>;
template class JTestGenerator<uint32_t>;
template class SweepGenerator<uint32_t>;